/*
 * userdev.c
 *
 *  Created on: Mar 28, 2013
 *      Author: bo se
 */

#include <linux/module.h>
#include <linux/rwsem.h>
#include <linux/pagemap.h>
#include <linux/highmem.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/cdev.h>
#include <linux/sched.h>
#include <asm/uaccess.h>
#include <asm/page.h>

#include "utok.h"

#define		DEV_NAME	"userdev"
#define		BUFFER_SIZE		( PAGE_SIZE * PAGE_COUNT )

static dev_t	first;
static unsigned int count		= 1;
static int major = 0, minor = 0;
static struct cdev *  user_cdev;
static struct page * pages[ PAGE_COUNT ];
static int p_count = 0;

ssize_t userdev_read ( struct file *  file, char __user *  data, size_t  size, loff_t *  offset )
{
	int how_much;
	int page = 0, page_offset = 0;
	void *  vaddress = NULL;

	if ( p_count <= 0 ) {
		printk( KERN_INFO "No data store on device " DEV_NAME );
		return 0;
	}

	how_much = min_t( size_t, size, PAGE_SIZE );

	if ( p_count * PAGE_SIZE <= *offset ) {
		return 0;
	}

	page = *offset / PAGE_SIZE;
	page_offset = *offset % PAGE_SIZE;

	how_much = min_t( size_t, how_much, PAGE_SIZE - page_offset );

	vaddress = kmap( pages[ page ] );

	copy_to_user( (void*)data, (void*)vaddress + page_offset, how_much );

	printk( KERN_INFO "Request: %lu; Read: %d; Offset: %lld; Vaddress: 0x%p;\n", size, how_much, *offset, vaddress );

	kunmap( pages[ page ] );

	*offset += how_much;

	return how_much;
}

static int ioctl_set_user_address( unsigned long  address )
{
	int j = 0;
	int	retval = 0;
	int count = 0;

	down_read( &current->mm->mmap_sem );
	count = get_user_pages(
			current, current->mm,
			address, PAGE_COUNT,
			1, 0,
			pages,	NULL
		);
	up_read( &current->mm->mmap_sem );

	if ( count != PAGE_COUNT ) {
		printk( KERN_WARNING "ioctl: error while pin pages user addr: %lx (%d).", address, count );
		if ( count > 0 ) {
			for ( j=0; j < count; ++j )
                page_cache_release( pages[ j ] );
		}
		retval = 0;
	} else {
		printk( KERN_INFO "ioctl: succes pin %d pages ( user address: %lx ).", count, address );
		retval = count;
	}
	return retval;
}

static void ioctl_unset_user_address( void )
{
	int i = 0;
	if ( p_count > 0 ) {
		for ( i = 0; i < p_count ; ++i ) {
			lock_page( pages[i] );
			set_page_dirty( pages[i] );
			unlock_page( pages[i] );
			page_cache_release( pages[i] );
		}
	}
	p_count = 0;
	return;
}

int userdev_release ( struct inode *inode, struct file *file )
{
	ioctl_unset_user_address();
	return 0;
}

long userdev_ioctl ( struct file * file, unsigned int cmd, unsigned long data )
{
	int		err = 0;
	long	retval = 0;

    if ( _IOC_TYPE(cmd) != USERDEV_IOC_MAGIC ) {
    	return -ENOTTY;
    }

    if ( _IOC_NR(cmd) > USERDEV_IOC_MAXNR ) {
    	return -ENOTTY;
    }

    if ( _IOC_DIR(cmd) & _IOC_READ ) {
            err		= !access_ok( VERIFY_WRITE, (void __user *)data, _IOC_SIZE(cmd) );
    } else if ( _IOC_DIR(cmd) & _IOC_WRITE ) {
            err		= !access_ok( VERIFY_READ, (void __user *)data, _IOC_SIZE(cmd) );
    }

    if (err) {
    	return -EFAULT;
    }

	switch (cmd) {
	case USERDEV_IO_SET_ADDRESS:
		p_count = ioctl_set_user_address( data );
		break;
	case USERDEV_IO_UNSET_ADDRESS:
		ioctl_unset_user_address();
		break;
	default:
		printk( KERN_DEBUG "IOCTL: no such command: %ud", cmd );
		retval	= -ENOTTY;
		break;
	};
	return retval;
}

static struct file_operations usercdev_fops = {
		.owner = THIS_MODULE,
		.unlocked_ioctl	= userdev_ioctl,
		.read = userdev_read,
		.release = userdev_release,
};

static int __init userdev_init( void )
{
#ifdef CHARDEV_MAJOR
	major = CHARDEV_MAJOR;
#else
	major = 700;
#endif
	first = MKDEV( major, minor );
	register_chrdev_region( first, count, DEV_NAME );
	user_cdev = cdev_alloc();
	cdev_init( user_cdev, &usercdev_fops );
	cdev_add( user_cdev, first, count );
	printk( KERN_WARNING "Device " DEV_NAME " registred.\n" );
	return 0;
}

static void __exit userdev_exit( void )
{
	ioctl_unset_user_address();
	cdev_del( user_cdev );
	unregister_chrdev_region( first, count );
	printk( KERN_WARNING "Device " DEV_NAME " unregistred.\n" );
	return;
}

module_init(userdev_init);
module_exit(userdev_exit);

MODULE_AUTHOR( "bo se" );
MODULE_DESCRIPTION( "Test of device direct I/O capability." );
MODULE_LICENSE( "GPL v2" );


